/* CSS reset */
html, body, div, span, h1, h2, h3, p, a, table, tr, th, td, ul, li {
    margin: 0;
    padding: 0;
    border: 0;
    outline: 0;
    font-size: 1em;
    vertical-align: baseline;
}
body, td, th {
    background-color: white;
    font-family: Arial, Helvetica, sans-serif;
    font-size: 11pt;
    line-height: 1.2em;
    color: black;
}
body {
    margin: 1.5em;
}
h1 {
    font-weight: bold;
    font-size: 1.6em;
    margin-top: 1em;
    margin-bottom: 0.6em;
}
h2 {
    font-weight: bold;
    font-size: 1.4em;
    margin-top: 1em;
    margin-bottom: 0.6em;
}
h3 {
    font-weight: bold;
    font-size: 1.2em;
    margin-top: 1em;
    margin-bottom: 0.6em;
}
p {
    margin-top: 0.5em;
}
h1:first-child, h2:first-child, h3:first-child, p:first-child {
    margin-top: 0;
}
a {
    color: #005090;
    text-decoration: none;
}
a:hover {
    text-decoration: underline;
}
a.linktarget {
    color: #000000;
    text-decoration: none;
}
table {
    margin-top: 0.4em;
    margin-left: 0.1em;
}
th {
    font-weight: bold;
    text-align: left;
    padding-top: 0.3em;
    padding-right: 0.8em;
}
th:last-child {
    padding-right: 0;
}
td {
    text-align: left;
    padding-top: 0.3em;
    padding-right: 0.8em;
}
td:last-child {
    padding-right: 0;
}

/* Index page */
ul.chisel-request-list {
    list-style: none;
    font-size: 1.2em;
    margin-top: 0.8em;
}
ul.chisel-request-list li ul {
    list-style: none;
}
ul.chisel-request-list li ul li {
    margin-top: 0.7em;
    margin-left: 0.2em;
}

/* Request page */
p.chisel-note {
    margin-top: 1em;
}
p.chisel-note ul {
    list-style: none;
    margin-top: 0.4em;
    margin-left: 0.8em;
}
ul.chisel-attr-list {
    list-style: none;
}
